import copy
from datetime import timedelta

from .params import Params
from .superdirt_client import SuperDirtClient
from .temporal_context import TemporalContext


class Pattern:
    def __init__(self, client: SuperDirtClient, params: dict) -> None:
        Event.validate(params)
        params_list = self.__prepare(params)
        events_list = self.__make_events_list(params_list)

        self.client = client
        self.events_list = events_list

    def __prepare(self, params: dict) -> list[dict]:
        longest = 0
        params2 = {}
        for k, v in params.items():
            if not isinstance(v, list):
                v = [v]
            if len(v) > longest:
                longest = len(v)
            params2[k] = v

        params_list = []
        for i in range(longest):
            params3 = {}
            for k, v in params2.items():
                params3[k] = v[i % len(v)]
            params_list.append(params3)
        return params_list

    def __make_events_list(
        self, params_list: list[dict]
    ) -> list[tuple[timedelta, list[dict]]]:
        events_list = []
        for params in params_list:
            delta = timedelta(seconds=params[Params.delta])
            events = Event.make_events(params)
            events_list.append((delta, events))
        return events_list

    def play(self, tctx: TemporalContext) -> None:
        for delta, events in self.events_list:
            for event in events:
                self.client.send(tctx, event)
            tctx.sleep(delta)


class Event:
    @classmethod
    def validate(cls, params: dict) -> None:
        if Params.s not in params and Params.sound not in params:
            raise ValueError(
                f"{Params.s} or {Params.sound} is required, params={params}"
            )
        if Params.delta not in params:
            raise ValueError(f"{Params.delta} is required, params={params}")

    @classmethod
    def make_events(cls, params: dict) -> list[dict]:
        # rest note
        s = params.get(Params.s)
        sound = params.get(Params.sound)
        if s is None and sound is None:
            return []

        # chord
        n = params.get(Params.n)
        chord = []
        if n is not None and isinstance(n, list):
            for note in params[Params.n]:
                note_params = copy.deepcopy(params)
                note_params[Params.n] = note
                chord.append(note_params)
            return chord

        # single note
        return [params]



def zmap(
    v: float, low_in: float, high_in: float, low_out: float, high_out: float
) -> float:
    v = min(max(v, low_in), high_in)
    percent = (v - low_in) / (high_in - low_in)
    return percent * (high_out - low_out) + low_out



from .chords import Chords
from .euclid import euclid
from .note import Note, PitchClass
from .pattern import Pattern
from .scale import Scale
from .scales import Scales
from .superdirt_client import SuperDirtClient
from .temporal_context import TemporalContext
from .utils import zmap

__all__ = [
    "Chords",
    "euclid",
    "Note",
    "PitchClass",
    "Pattern",
    "Scale",
    "Scales",
    "SuperDirtClient",
    "TemporalContext",
    "zmap",
]



def euclid(k: int, n: int, r: int = 0) -> str:
    if not n >= 2:
        raise ValueError(f"n must be greater than 2: n={n}")
    if not 0 < k < n:
        raise ValueError(f"k must be 0 < k < n: k={k}, n={n}")

    xss, yss = [["x"] for _ in range(k)], [["."] for _ in range(n - k)]
    xss, yss = _euclid(xss, yss)
    xs = sum(xss + yss, [])
    r %= len(xs)
    xs = xs[r:] + xs[:r]
    return "".join(xs)


def _euclid(
    xss: list[list[str]], yss: list[list[str]]
) -> tuple[list[list[str]], list[list[str]]]:
    size = min(len(xss), len(yss))
    for i in range(size):
        xss[i] += yss[i]
    xss, yss = xss[:size], xss[size:] + yss[size:]
    if len(yss) < 2:
        return xss, yss
    return _euclid(xss, yss)



from .note import Note, PitchClass
from .scales import Scales


class Scale:
    def __init__(self, root: PitchClass, scale: list[int] = Scales.chromatic) -> None:
        self.root = root
        self.scale = scale

    def __map(self, degree: int, octave: int) -> Note:
        octave += degree // len(self.scale)
        degree = degree % len(self.scale)
        note = Note(self.root, octave).transpose(self.scale[degree])
        return note

    def bind(self, degrees: list, octave: int = 0) -> list:
        ret: list = []
        for degree in degrees:
            if degree is None:
                # rest note
                ret.append(None)
            elif not isinstance(degree, list):
                # single note
                ret.append(self.__map(degree=degree, octave=octave).midi_number)
            else:
                # chord
                ret.append(self.bind(degrees=degree, octave=octave))
        return ret

    def degrees(self) -> list[int]:
        return sorted([(x + self.root) % 12 for x in self.scale])



from __future__ import annotations

from enum import IntEnum


class PitchClass(IntEnum):
    C = 0
    Cs = 1
    D = 2
    Ds = 3
    E = 4
    F = 5
    Fs = 6
    G = 7
    Gs = 8
    A = 9
    As = 10
    B = 11


class Note:
    def __init__(self, pc: PitchClass, octave: int) -> None:
        midi_number = self.__to_midi_number(pc, octave)

        self.pc = pc
        self.octave = octave
        self.midi_number = midi_number

    def __repr__(self) -> str:
        class_name = type(self).__name__
        return f"{class_name}(pc={self.pc!r}, octave={self.octave!r}, midi_number={self.midi_number!r})"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Note):
            return NotImplemented
        return self.midi_number == other.midi_number

    def __lt__(self, other: Note) -> bool:
        return self.midi_number < other.midi_number

    def __le__(self, other: Note) -> bool:
        return self.midi_number <= other.midi_number

    def __hash__(self) -> int:
        return hash(self.midi_number)

    def transpose(self, n: int) -> Note:
        midi_number = self.midi_number + n
        pc, octave = self.__from_midi_number(midi_number)
        return Note(pc, octave)

    @classmethod
    def __to_midi_number(cls, pc: PitchClass, octave: int) -> int:
        return pc + octave * 12

    @classmethod
    def __from_midi_number(cls, midi_number: int) -> tuple[PitchClass, int]:
        pc = PitchClass(midi_number % 12)
        octave = int(midi_number / 12)
        return pc, octave



from __future__ import annotations

import copy
import time
from datetime import datetime, timedelta


class TemporalContext:
    def __init__(self, dryrun: bool = False) -> None:
        init_vtime = datetime.now()
        self.__init_vtime = init_vtime
        self.__vtime = init_vtime
        self.__dryrun = dryrun

    def elapsed_time(self) -> timedelta:
        return self.__vtime - self.__init_vtime

    def now(self) -> datetime:
        return self.__vtime

    def set_now(self, now: datetime) -> None:
        self.__vtime = now

    def __sleep(self) -> None:
        if self.is_dryrun():
            return
        delta_sec = (self.__vtime - datetime.now()).total_seconds()
        if delta_sec > 0:
            time.sleep(delta_sec)

    def sleep(self, delta: timedelta) -> None:
        self.__vtime += delta
        self.__sleep()

    def sleep_until(self, until: datetime) -> None:
        self.__vtime = until
        self.__sleep()

    def is_dryrun(self) -> bool:
        return self.__dryrun

    def clone(self) -> TemporalContext:
        return copy.deepcopy(self)



import logging

from pythonosc import osc_bundle_builder, osc_message_builder, udp_client

from .temporal_context import TemporalContext

logger = logging.getLogger(__name__)


class OscClient:
    def __init__(self, address: str, port: int) -> None:
        self.__client = udp_client.SimpleUDPClient(address, port)

    def send(self, address: str, args: dict, timestamp: float) -> None:
        bundle = osc_bundle_builder.OscBundleBuilder(timestamp)  # type: ignore
        msg = osc_message_builder.OscMessageBuilder(address=address)
        for k, v in args.items():
            msg.add_arg(k)
            msg.add_arg(v)
        bundle.add_content(msg.build())  # type: ignore
        self.__client.send(bundle.build())


class SuperDirtClient:
    ADDRESS: str = "/dirt/play"

    def __init__(
        self, address: str = "127.0.0.1", port: int = 57120, delay: float = 1.0
    ) -> None:
        self.__client = OscClient(address, port)
        self.__delay = delay

    def send(self, tctx: TemporalContext, event: dict) -> None:
        if tctx.is_dryrun():
            return
        logger.debug(event)
        self.__client.send(
            address=self.ADDRESS,
            args=event,
            timestamp=tctx.now().timestamp() + self.__delay,
        )
